/**********************************************************************
 * Software Copyright Licensing Disclaimer
 *
 * This software module was originally developed by contributors to the
 * course of the development of ISO/IEC 14496-10 for reference purposes
 * and its performance may not have been optimized.  This software
 * module is an implementation of one or more tools as specified by
 * ISO/IEC 14496-10.  ISO/IEC gives users free license to this software
 * module or modifications thereof. Those intending to use this software
 * module in products are advised that its use may infringe existing
 * patents.  ISO/IEC have no liability for use of this software module
 * or modifications thereof.  The original contributors retain full
 * rights to modify and use the code for their own purposes, and to
 * assign or donate the code to third-parties.
 *
 * This copyright notice must be included in all copies or derivative
 * works.  Copyright (c) ISO/IEC 2004, 2005, 2006, 2007, 2008.
 **********************************************************************/

/*!
*************************************************************************************
* \file  c3d_generic_filters.h
*
* \brief 
*        MFC SDK  Generic Downsample filtering functions
*
* \author 
*    Main contributors
*        - Hariharan Ganapathy   (hari.ganapathy@dolby.com)
*
*
*************************************************************************************
*/


#ifndef _C3D_GENERIC_FILTERS
#define _C3D_GENERIC_FILTERS

/**
 * \fn void vertical_filter_and_downsample(
 *                                     const  RPUFilter       *p_flt,
 *                                     const  RPUFilterParams *p_filter_params,
 *                                     const  img_pel_t       *p_in,
 *                                            uint16_t        u16_input_stride,
 *                                            img_pel_t       *p_out,
 *                                            uint16_t        u16_output_stride, 
 *                                    )
 *
 * \brief vertical_filter_and_downsample.
 *
 * \details * \details Vertically downsample input partition by seperately processing boundary and center pixels.
 *  Independant of packed or unpacked Data
 *
 * \param[in]    p_flt                    Pointer to filter definition. This is typecast to 1D filter.
 * \param[in]    p_filter_params            Pointer to filter params. This contains spatial information about the center and 
 *                                        boundary regions of the partition. 
 * \param[in]    p_in                    Pointer to the input buffer. p_in points to the start of partition.
 * \param[in]    u16_input_stride        Input buffer stride.
 * \param[in]    p_out                    Pointer to the output buffer.
 * \param[in]    u16_output_stride        Output buffer stride. 
 *
 * \return None
 */
void vertical_filter_and_downsample(                                    
                              const RPUFilter    *p_flt,
                              const RPUFilterParams    *p_filter_params,
                              const img_pel_t        *p_in,
                                    uint16_t        u16_input_stride,
                                    img_pel_t        *p_out,
                                    uint16_t        u16_output_stride
                             );

/**
 * \fn void horizontal_filter_and_downsample(
 *                                     const  RPUFilter    *p_flt,
 *                                     const  RPUFilterParams    *p_filter_params,
 *                                     const  img_pel_t        *p_in,
 *                                            uint16_t        u16_input_stride,
 *                                            img_pel_t        *p_out,
 *                                            uint16_t        u16_output_stride
 *                                    )
 *
 * \brief horizontal_filter_and_downsample.
 *
 * \details Filters the input partition with the given 1D filter to obtain the output. It is assumed that the filter is
 * symmetric. The filter processes one line at a time. It calls functions to filter the left, center and right columns.
 * The left and right columns have leading and trailing edges, respectively. Duplication of data is needed for 
 * processing these columns. However, the center column doesn't need any duplication. It has all the required data.
 *
 * \param[in]    p_flt                    Pointer to filter definition. This is typecast to 1D filter.
 * \param[in]    p_filter_params            Pointer to filter params. This contains spatial information about the center and 
 *                                        boundary regions of the partition. 
 * \param[in]    p_in                    Pointer to the input buffer. p_in points to the start of partition.
 * \param[in]    u16_input_stride        Input buffer stride.
 * \param[in]    p_out                    Pointer to the output buffer.
 * \param[in]    u16_output_stride        Output buffer stride. 
 *
 * \return None
 */
void horizontal_filter_and_downsample(                                    
                              const RPUFilter    *p_flt,
                              const RPUFilterParams    *p_filter_params,
                              const img_pel_t        *p_in,
                                    uint16_t        u16_input_stride,
                                    img_pel_t        *p_out,
                                    uint16_t        u16_output_stride,
                                    uint8_t            u8_packed_UV
                             );

#endif /* _C3D_GENERIC_FILTERS */
